//======================================================== IBJhdr.c
//
// Create header IBJtbl.h from user definitions in IBJtbl.txt
// ==========================================================
//
// Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999-2000
// email: lj@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
//
// 98-07-31 lj
// 99-01-19 lj  casting to (double*) corrected
// 99-01-21 lj  pointer type added
// 2000-07-28 uj new program arguments
//
//==================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "IBJprm.h"

#define  NMAX  200
#define  NSC   2

typedef struct {
  char name[40], dsc[120], frm[40], dfl[40], ref[40];
  int type;
} INPREC;

static INPREC *Itab;
static int Nmax = NMAX;
static char Path[256], Pgm[80];

static int trim( char *pc, int n )
{
  char *p1, *p2;
  for (p2=pc; (*p2); p2++)  if (*p2 > ' ')  break;
  if (!*p2) {
    *pc = 0;
    return 0;
  }
  for (p1=pc; (*p2); ) *p1++ = *p2++;
  *p1-- = 0;
  for ( ; p1>pc && *p1<=' '; p1--)  *p1 = 0;
  if (strlen(pc) >= n) {
    pc[n-1] = 0;
    return 1;
  }
  return 0;
}


static int ReadList( char *name )
{
  char fn[256], buf[256], *pc;
  int n;
  FILE *f;
  strcpy(fn, Path);
  strcat(fn, name);
  f = fopen(fn, "rb");
  if (!f) {
    printf("can't open file %s\n", fn);
    exit(2);
  }
  Itab = calloc(Nmax, sizeof(INPREC));
  if (!Itab) {
    printf("can't allocate input table\n");
    exit(3);
  }
  n = 0;
  while (fgets(buf, 256, f)) {
    if (n >= Nmax) {
      printf("at most %d input lines allowed\n", Nmax);
      exit(4);
    }
    if (*buf == '-')  continue;
    if (*buf == ':') {
      sscanf(buf+1, "%s", Pgm);
      continue;
    }
    pc = strtok(buf, ";");  if ((!pc) || (*pc<=' '))  continue;
    trim(pc, 40);
    strcpy(Itab[n].name, pc);
    pc = strtok(NULL, ";");  if (!pc)  break;
    trim(pc, 40);
    strcpy(Itab[n].frm, pc);
    pc = strtok(NULL, ";");  if (!pc)  break;
    trim(pc, 40);
    strcpy(Itab[n].dfl, pc);
    pc = strtok(NULL, ";");  if (!pc)  break;
    trim(pc, 40);
    strcpy(Itab[n].ref, pc);
    pc = strtok(NULL, ";");  if (!pc)  break;
    trim(pc, 120);
    strcpy(Itab[n].dsc, pc);
    n++;
  }
  if (!pc) {
    printf("can't parse line %d\n", n);
    exit(5);
  }
  fclose(f);
  return n;
}

static int WriteHeader( int n, char *name )
{
  int i, j, size, type, len, k, l;
  char *pc, fn[256], ifrm[40], ofrm[40], xfrm[48], dfl[40], ref[40];
  char vname[40], mname[40];
  char sifrm[40], sxfrm[40], sdfl[40], sref[40];
  FILE *f;
  strcpy(fn, Path);
  strcat(fn, name);
  f = fopen(fn, "wb");
  if (!f) {
    printf("can't open file %s\n", fn);
    exit(10);
  }
  fprintf(f, "// Copyright Dr. Lutz Janicke, Dunum, Germany, 1998-2000\n\n");
  fprintf(f, "// This is the header IBJtab.h to be used by the program \"%s\".\n", Pgm);
  fprintf(f, "// It defines all parameters whose values can be given by the user.\n");
  fprintf(f, "// It has to be included by all modules that use these values.\n\n");
  fprintf(f, "// Each parameter is defined in a record of the struct PrmTab.\n");
  fprintf(f, "// The name of the record is the same as the name of the input\n");
  fprintf(f, "// parameter (small letters). The record contains the name, a\n");
  fprintf(f, "// short description, input and output format and the default\n");
  fprintf(f, "// value as a string. For arrays also the number of elements or\n");
  fprintf(f, "// a reference to a parameter is given, that represents this\n");
  fprintf(f, "// number. For details see the header IBJprm.h.\n\n");
  fprintf(f, "#ifndef  IBJPRM_INCLUDE\n");
  fprintf(f, "    #include \"IBJprm.h\"  // The typedef of PRMTABREC is needed\n");
  fprintf(f, "#endif\n\n");
  fprintf(f, "#ifndef  IBJPRM_SOURCE\n");
  fprintf(f, "    extern  // Only IBJPRM saves the data of this structure\n");
  fprintf(f, "#endif\n");
  fprintf(f, "struct { PRMTABREC  // The names of the input parameters:\n");
  for (i=0; i<n; i++) {
    strcpy(vname, Itab[i].name);
    if (*vname == '*')
      for (pc=vname; (*pc); pc++)
        if (*pc == '*')  *pc = '_';
    fprintf(f, " %s,", vname);
    if (!((i+1)%20))  fprintf(f, "\n");
  }
  fprintf(f, " end_of_table; } PrmTab\n");
  fprintf(f, "#ifdef  IBJPRM_SOURCE\n");
  fprintf(f, "//      The initial values are defined in IBJPRM\n");
  fprintf(f, " = {\n");
  for (i=0; i<n; i++) {
    strcpy(ofrm, Itab[i].frm);
    l = strlen(ofrm);
    len = -1;
    sscanf(ofrm, "%d", &len);
    switch (ofrm[l-1]) {
      case 'd': type = PRM_ISINT;
            strcpy(ifrm, "%d");
            break;
      case 'e':
      case 'f': if (l>1 && ofrm[l-2]=='l')  type = PRM_ISDOUBLE;
            else                        type = PRM_ISFLOAT;
            break;
      case 'p': type = PRM_ISPOINTER;
                break;
      case 's': type = PRM_ISSTRING;
                break;
      default:  printf("unknown format %s in definition of %s\n",
              ofrm, Itab[i].name);
                exit(11);
    }
    switch (type) {
      case PRM_ISINT:   strcpy(ifrm, "%d");
                size = sizeof(int);
                          break;
      case PRM_ISFLOAT:   strcpy(ifrm, "%f");
                size = sizeof(float);
                          break;
      case PRM_ISDOUBLE:  strcpy(ifrm, "%lf");
                size = sizeof(double);
                          break;
      case PRM_ISPOINTER: strcpy(ifrm, "%p");
                size = sizeof(void*);
                          break;
      case PRM_ISSTRING:  size = (len < 0) ? -len : len;
                if (!size)  size = 79;
                strcpy(ifrm, "%s");
                break;
      default:            printf("unknown type %d in definition of %s\n",
                  type, Itab[i].name);
                          exit(12);
    }
    strcpy(xfrm, "");
    if (type == PRM_ISSTRING) {
      strcat(xfrm, "\\\"%");
      strcat(xfrm, "s");      // strcat(xfrm, ofrm);
      strcat(xfrm, "\\\"");
    }
    else {
      strcat(xfrm, "%");
      strcat(xfrm, ofrm);
    }
    strcpy(dfl, Itab[i].dfl);
    strcpy(ref, Itab[i].ref);
    k = -1;
    if (!*ref)  k = 0;
    else  sscanf(ref, "%d", &k);
    if (k > 0)  type |= PRM_ISARRAY;
    else if (k < 0) {
      for (j=0; j<i; j++)
          if (!strncmp(Itab[j].name, ref, NSC))  break;
      if (j >= i) {
        printf("unknown reference %s in definition of %s\n",
                ref, Itab[i].name);
        exit(13);
      }
      type |= PRM_ISARRAY;
    }
    Itab[i].type = type;
    sprintf(sifrm, "\"%s\"", ifrm);
    sprintf(sxfrm, "\" %s\"", xfrm);
    sprintf(sdfl, "\"%s\"", dfl);
    sprintf(sref, "\"%s\"", ref);
    fprintf(f, " { \"%s\", %5s, %10s, %8s, %6s, %3d, %2d, 0x%03x, 0, \"%s\" },\n",
      Itab[i].name, sifrm, sxfrm, sdfl, sref, size, k, type, Itab[i].dsc);
  }
  fprintf(f, " { \"**\", NULL, NULL, NULL, NULL, 0, 0, 0, NULL } }\n");
  fprintf(f, "#endif\n");
  fprintf(f, ";\n//  Macros for easier reference:\n\n");
  for (i=0; i<n; i++) {
    strcpy(vname, Itab[i].name);
    if (*vname == '*')  continue;
    strcpy(mname, vname);
    for (j=strlen(mname)-1; j>=0; j--)  mname[j] = toupper(mname[j]);
    type = Itab[i].type;
    if (type & PRM_ISARRAY)
      switch (type & PRM_ISANY) {
        case PRM_ISINT:
          fprintf(f, "#define %s(i) ((int*)PrmTab.%s.u.p)[i]\n", mname, vname);
          break;
        case PRM_ISFLOAT:
          fprintf(f, "#define %s(i) ((float*)PrmTab.%s.u.p)[i]\n", mname, vname);
          break;
        case PRM_ISDOUBLE:          //-99-01-19
          fprintf(f, "#define %s(i) ((double*)PrmTab.%s.u.p)[i]\n", mname, vname);
          break;
        case PRM_ISPOINTER:         //-99-01-21
          fprintf(f, "#define %s(i) ((void**)PrmTab.%s.u.p)[i]\n", mname, vname);
          break;
        case PRM_ISSTRING:
          fprintf(f, "#define %s(i) ((char**)PrmTab.%s.u.p)[i]\n", mname, vname);
          break;
        default: ;
      }
    else
      switch (type & PRM_ISANY) {
        case PRM_ISINT:
          fprintf(f, "#define  %s  PrmTab.%s.u.i\n", mname, vname);
          break;
        case PRM_ISFLOAT:
          fprintf(f, "#define  %s  PrmTab.%s.u.f\n", mname, vname);
          break;
        case PRM_ISDOUBLE:
          fprintf(f, "#define  %s  PrmTab.%s.u.d\n", mname, vname);
          break;
        case PRM_ISSTRING:
          fprintf(f, "#define  %s  (char*)PrmTab.%s.u.p\n", mname, vname);
          break;
        default: ;
      }
  }

  fclose(f);
  return 0;
}


int main( int argc, char *argv[] )
{
  int l, n;
  if (argc < 3) {
    printf("usage: IBJhdr <path> <parameter file>\n");
    exit(0);
  }
  strcpy(Path, argv[1]);
  l = strlen(Path);
  for (l=0; (Path[l]); l++)
    if (Path[l] == '\\')  Path[l] = '/';
  if (Path[l-1] != '/')  strcat(Path, "/");
  n = ReadList(argv[2]);
  n = WriteHeader(n, "IBJtbl.h");
  free(Itab);
  printf("header generated\n");
  return 0;
}

