//======================================================== IBJmsg.c
//
// Message Routines for IBJ-programs
// =================================
//
// Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999-2000
// email: lj@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
//
// last change: 98-07-31  lj
//
//==================================================================

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>

#include "IBJmsg.h"

char *MsgVersion = "1.1.a";

FILE *MsgFile;
int MsgCode, MsgQuiet, MsgVerbose;
int MsgTranslate = 1;
int MsgBufLen = 4000;
//
// static char *eMODn = "IBJmsg";
//
/*===================================================================== MsgLow
*/
void MsgLow(    /* convert to lower case        */
char *s )       /* string to be converted       */
  {
  while (*s) {
    *s = tolower(*s);
    s++;
    }
  }

/*===================================================================== MsgTok
*/
char *MsgTok(   /* non-destructive tokenisation */
char *start,    /* begin of scanning            */
char **pnext,   /* start of next scan           */
char *sep,      /* separator characters         */
int *pl )       /* number of usable characters  */
  {             /* RETURN: start of token       */
  int l, l1;
  if ((!start) || (!sep))  return NULL;
  while (*start && strchr(sep, *start))  start++;
  if (!*start)  return NULL;
  l1 = 0;
  if (*start == '\"') {
    start++;
    for (l=0; ; l++) {
      if (!start[l])  break;
      if (start[l] == '\"') { l1 = 1;  break; }
      }
    }
  else {
    for (l=0; ; l++)  if (!start[l] || strchr(sep, start[l]))  break;
    }
  if (pnext)  *pnext = start + l + l1;
  if (pl)  *pl = l;
  return start;
  }

//====================================================================== vMsg
int vMsg(       /* print message on screen and into logfile     */
char *format,   /* printing format                              */
... )           /* parameter                                    */
{
  va_list argptr;
  int verbose, n;
  static char *s, *pc;
  static char DSTcode[] = { 0x84,0x94,0x81,0x8E,0x99,0x9A,0xE6,0xE1,0xFD,0xFC,0x00 };
  static char SRCcode[] = { 0xE4,0xF6,0xFC,0xC4,0xD6,0xDC,0xB5,0xDF,0xB2,0xB3,0x00 };
  int i, l;
  if (!format) {
    if (s)  FREE(s);
    s = NULL;
    return 0;
  }
  if (!s) {
    s = ALLOC(MsgBufLen);
    if (!s) {
      printf("*** can't allocate message buffer (%d bytes)!\n", MsgBufLen);
      fflush(stdout);
      exit(1);
    }
  }
  verbose = (MsgCode < 0) ? 1 : MsgVerbose;
  if (MsgQuiet<=0 && verbose) {
    va_start(argptr, format);
    n = _vbprintf(s, MsgBufLen-1, format, argptr);
    if (n < 0)  sprintf(s, "--- message buffer overflow!\n");
    va_end(argptr);
    if (MsgTranslate) {
      l = strlen(s);
      for (i=0; i<l; i++) {
        pc = strchr(SRCcode, s[i]);
        if (pc)  s[i] = DSTcode[pc-SRCcode];
      }
    }
    printf("%s", s);  fflush(stdout);
  }
  va_start(argptr, format);
  if (MsgFile && verbose)  vfprintf(MsgFile, format, argptr);
  va_end(argptr);
  return MsgCode;
}


/*=============================================================== MsgDateVal
*/

static void trim( char * s )
  {
  char *p, *q;
  for (p=s, q=s; *q==' '; q++);
  for ( ; (*q); p++, q++)  *p = *q;
  *p = 0;
  for (p--; p>=s; p--)
      if (*p == ' ')  *p = 0;
      else  break;
  }

double MsgDateVal(      /* converts date string into numerical value    */
char * ss )             /* string of the form yy-mm-dd'hh:mm:ss         */
  {                     /* RETURN: days after 1899-12-30'00:00:00 +1.e6 */
  static int dayspassed[] = {
        0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 };
  char s[80], *pc, *pd, *pt;
  int year, month, day, days;
  int hour, minute, second, n;
  double mydate;
  strcpy(s, ss);  trim(s);
  if (!strcmp(s, "-inf"))  return -9999999;
  if (!strcmp(s, "+inf"))  return 9999999;
  pc = strchr(s, ' ');
  if (!pc)  pc = strchr(s, '\'');
  if (!pc)  pc = strchr(s, '.');
  if (pc) {
      *pc = 0;
      pt = pc+1;
      pd = s; }
  else {
      if (strchr(s, '-')) {
          pd = s;
          pt = ""; }
      else {
          pd = "";
          pt = s; }
      }
  pc = strchr(pd, '-');
  mydate = 0;
  if (pc) {
      *pc++ = 0;
      year = atoi(pd);
      pd = pc;
      pc = strchr(pd, '-');
      if (pc) {
          *pc++ = 0;
          month = atoi(pd);
          day = atoi(pc); }
      else {
          month = year;
          year = 0;
          day = atoi(pd); }
      if (year > 1000)  year = year - 1900;
      if (year < 0)  year = 0;
      if (month < 0)  month = 0;
      if (month > 12)  month = 12;
      days = 0;
      if (month > 0) {
          days += 1 + dayspassed[month-1];
          if (year > 0)  days += 365*year + (year-1)/4;
          if ((year>0) && (year%4==0) && (month>2))  days++; }
      days += day;
      mydate = 1000000 + days;
      }
  else  sscanf(pd, "%lf", &mydate);
  n = sscanf(pt, "%d:%d:%d", &hour, &minute, &second);
  if (n < 3) { second=minute; minute=hour; hour=0; }
  if (n < 2) { second=minute; minute=0; }
  if (n < 1) { second=0; }
  mydate += (second + 60*(minute + 60*hour))/(24.0*3600.0);
  return mydate;
  }

char *MsgDateString(    /* converts date value into string      */
double d )              /* days after 1899-12-30'00:00:00 +1.e6 */
  {                     /* RETURN: string yy-mm-dd'hh:mm:ss     */
  static char dstr[80];
  static int dayspassed[] = {
        0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 };
  int seconds, second, minute, hour;
  int days, day, month, year;
  int day1, day2, add;
  if (d < 0)  return "-inf";
  if (d >= 9999999)  return "+inf";
  days = floor(d);
  seconds = (d - days)*24*3600 + 0.5;
  second = seconds % 60;
  minute = (seconds/60) % 60;
  hour = seconds/3600;
  if (days == 0) {
      sprintf(dstr, "%02d:%02d:%02d", hour, minute, second);
      return dstr; }
  if (days < 1000002) {
      sprintf(dstr, "%d'%02d:%02d:%02d", days, hour, minute, second);
      return dstr; }
  days -= 1000000;
  day1 = 2;
  for (year=0; year<200; year++) {
      day2 = day1 + 365;
      if ((year) && (year%4==0)) day2++;
      if (days < day2)  break;
      day1 = day2; }
  days -= day1;
  add = ((year) && (year%4==0)) ? 1 : 0;
  for (month=1; month<=12; month++)
      if (days < dayspassed[month]+add*(month>1))  break;
  day = 1 + days - (dayspassed[month-1]+add*(month>2));
  if (year >= 100)  year += 1900;
  sprintf(dstr, "%02d-%02d-%02d'%02d:%02d:%02d",
    year, month, day, hour, minute, second);
  return dstr;
  }

double MsgDateSeconds( double t1, double t2 )
  {
  return (t2-t1)*24*3600;
  }


#ifdef  MAIN  /*#####################################################*/

int main()
  {
  char s[80], t[80];
  double d;
  for (;;) {
      fputs(":", stdout);  gets(s);
      if (!*s) break;
      trim(s);
      d = MsgDateVal(s);
      strcpy(t, MsgDateString(d));
      printf("%15.5lf  >%s<\n", d, t);
  }
  return 0;
  }
#endif
/*==========================================================================*/
